# Import libraries
import paramiko
import sys
import time 
from termcolor import colored
import colorama
colorama.init()

class SSHClient:

  def __init__(self, username, ip, password, keyFile=''):
    print(f'Attempting to SSH to {ip}...')
    self.ip = ip 
    self.username = username
    self.password = password 
    self.keyFile = keyFile
    self.client = paramiko.SSHClient()
    self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    self._Connect()

  def __del__(self):
    self._Disconnect()

  def _Connect(self):
    try:
      self.client.connect(self.ip, 22, self.username, self.password, key_filename=self.keyFile)
      print('Connected!')
    except:
      print(colored('Connection failed, trying again...', 'yellow'))
      self._Connect()

  def _Disconnect(self):
    try:
      self.client.close()
      print('Closed SSH connection!')
    except:
      print(colored(f'Failed to disconnect from {self.ip}!', 'red'))
      print(sys.exc_info()[0])

  def Call(self, command, printResult=False):
    stdin, stdout, stderr = self.client.exec_command(command)
    for line in iter(lambda: stdout.readline(2048), ''):
      if printResult:
        print(line, end='')
    return stdin, stdout, stderr

  def OpenSFTP(self):
    return self.client.open_sftp()
  
  def CloseSFTP(self, sftpClient):
    sftpClient.close()
    return 

  def CheckFileExistsOnRemote(self, path):
    sftpClient = self.OpenSFTP()
    try:
      sftpClient.stat(path)
      return True 
    except:
      return False


# Imports
import boto3
import time 
import sys
import subprocess
from termcolor import colored
import colorama
colorama.init()

class AWSInstance:

  def __init__(self, region, id, shutdownOnDestroy=True):
    self.session = boto3.session.Session(region_name=region)
    self.client = self.session.client('ec2')
    self.id = id
    self.shutdownOnDestroy = shutdownOnDestroy
    self.ip = None
    self.online = self.CheckRunning()
  
  def Startup(self):
    if self.CheckRunning():
      print(colored('Remote server is already running!', 'yellow'))
      self.online = True
      return True
    else:
      print("Booting up remote server...")
      try:
        self.client.start_instances(InstanceIds=[self.id])
      except:
        print(colored('Remote server failed to start! It may be in an unstable state, try again in a few seconds.', 'red'))
        self.online = False
        return False
      print("Waiting for remote server to be set up before continuing...")
      while not self.CheckRunning():
        time.sleep(0.1)
      print('Remote server is online!')
      self.online = True
      return True

  def Shutdown(self):
    print("Shutting down remote server...")
    self.client.stop_instances(InstanceIds=[self.id])
    while not self.CheckStopped():
      time.sleep(0.1)
    print("Remote server has been shut down!")

  def CheckRunning(self):
    resp = self.client.describe_instance_status(InstanceIds=[self.id])
    if len(resp['InstanceStatuses']) == 0:
      return False
    return resp['InstanceStatuses'][0]['InstanceState']['Name'] == 'running'

  def CheckStopped(self):
    resp = self.client.describe_instance_status(InstanceIds=[self.id])
    if len(resp['InstanceStatuses']) == 0:
      return True
    return resp['InstanceStatuses'][0]['InstanceState']['Name'] == 'stopped'

  def GetInstanceIP(self):
    if self.ip == None:
      self.ip = self.client.describe_instances(InstanceIds=[self.id])['Reservations'][0]['Instances'][0]['NetworkInterfaces'][0]['PrivateIpAddresses'][0]['Association']['PublicIp']
    return self.ip

  def PingInstance(self):
    command = ['ping', '-n', '1', self.GetInstanceIP()]
    if subprocess.call(command, stdout=subprocess.PIPE) == 0:
      return True
    else:
      return False

  def ExecutePowerShellScript(self, scriptPath):
    ssmClient = boto3.client('ssm')
    try:
      scriptLines = open(scriptPath, 'r').readlines()
    except IOError:
      print(colored('Script was not found!', 'red'))
      return
    params = {'commands': scriptLines}
    resp = ssmClient.send_command(DocumentName='AWS-RunPowerShellScript', InstanceIds=[self.id], Parameters=params)
    commandID = resp['Command']['CommandId']
    output = {'ResponseCode': -1}
    while output['ResponseCode'] != 0:
      try:
        output = ssmClient.get_command_invocation(CommandId=commandID, InstanceId=self.id)
      except:
        pass
    return output['StandardOutputContent']


# Imports
import os
import json
import shutil 
import subprocess
import stat 
import ssh
import time
from termcolor import colored
import colorama
colorama.init()

class GMBuilder:

  def __init__(self, config, sshClient, verbose):

    def LoadOptions(config):
      self._PrintWrapper('Loading user configured options...')
      config['projectName'] = os.path.basename(config['yypPath']).replace('.yyp', '')
      config['remoteProjectPath'] = 'C:/users/buildManager/AppData/Local/lazybuild/Input/Project'
      return config

    self._sshClient = sshClient
    self._verbose = verbose
    configData = LoadOptions(config)
    self.outputFolder = 'C:/users/buildManager/AppData/Local/lazybuild'
    self.cacheFolder = 'C:/Users/buildManager/AppData/Roaming/GameMakerStudio2/Cache/GMS2CACHE'
    self._yoyoID = configData['yoyoID']
    self._runtimeVersion = configData['runtimeVersion']
    self._steamSDKPath = configData['steamSDKPath']
    self._remoteProjectPath = configData['remoteProjectPath']
    self._projectName = configData['projectName']
    self._configuration = configData['configuration']
    self._gitBranch = configData['gitBranch']
    self._gitURL = configData['gitURL']
    self._gitUsername = configData['gitUsername']
    self._gitPassword = configData['gitPassword']
    self._useCache = True

  def _PrintWrapper(self, message):
    if self._verbose:
      print(message)

  def _SanatizeOptionsDictionary(self, options):
    self._PrintWrapper('Populating file with user data...')
    for key, value in options.items():
      value = value.replace('{BUILD_PATH}', self.outputFolder)
      value = value.replace('{YOYO_ID}', self._yoyoID)
      value = value.replace('{RUNTIME_VERSION}', self._runtimeVersion)
      value = value.replace('{PROJECT_NAME}', self._projectName)
      value = value.replace('{USERNAME}', self._gitUsername)
      value = value.replace('{CONFIG}', self._configuration)
      options[key] = value
    return options

  def _GenerateBuildFile(self):
    self._PrintWrapper('Generating build.bff...')
    buildTemplate = json.load(open('resources/options/build.bff', 'r'))
    buildTemplate = self._SanatizeOptionsDictionary(buildTemplate)
    return buildTemplate

  def _GenerateMacrosFile(self):
    self._PrintWrapper('Generating macros.json...')
    macrosTemplate = json.load(open('resources/options/macros.json', 'r'))
    macrosTemplate = self._SanatizeOptionsDictionary(macrosTemplate)
    return macrosTemplate

  def DumpOptionFiles(self):
    self._PrintWrapper('Creating build option files on remote...')
    sftpClient = self._sshClient.OpenSFTP()
    sftpClient.put('resources/options/targetoptions.json', os.path.join(self.outputFolder, 'Input', 'Options', 'targetoptions.json'))
    sftpClient.put('resources/options/preferences.json', os.path.join(self.outputFolder, 'Input', 'Options', 'preferences.json'))
    json.dump({'steamsdk_path': self._steamSDKPath}, open('temp', 'w+'))
    sftpClient.put('temp', os.path.join(self.outputFolder, 'Input', 'Options', 'steam_options.yy'))
    json.dump(self._GenerateBuildFile(), open('temp', 'w+'), indent=4)
    sftpClient.put('temp', os.path.join(self.outputFolder, 'Input', 'Options', 'build.bff'))
    json.dump(self._GenerateMacrosFile(), open('temp', 'w+'), indent=4)
    sftpClient.put('temp', os.path.join(self.outputFolder, 'Input', 'Options', 'macros.json'))
    os.remove('temp')
    
  def CreateBuildFolders(self, clearCache=False):

    def PurgeDirectory(path):
      self._sshClient.Call(f'rmdir /S /Q "{path}"')
      self._sshClient.Call(f'mkdir "{path}"')

    self._PrintWrapper('Creating build folder architecture...')
    self._sshClient.Call(f'mkdir "{self.outputFolder}"') # No purge to not deleted project, but it has to exist!
    PurgeDirectory(os.path.join(self.outputFolder, 'Output', 'GameZip'))
    PurgeDirectory(os.path.join(self.outputFolder, 'Output', 'GameFiles'))
    PurgeDirectory(os.path.join(self.outputFolder, 'Input', 'Options'))
    if clearCache:
      self._PrintWrapper('Clearing the cache...')
      PurgeDirectory(os.path.join(self.cacheFolder))
      
  def UpdateProject(self):

    self._PrintWrapper('Opening SFTP connection...')
    sftpClient = self._sshClient.OpenSFTP()
    
    # Clone vs Pull
    try:
      sftpClient.stat(os.path.join(self._remoteProjectPath, '.git')) # path test
      self._PrintWrapper('Found project repo!')
      self._PrintWrapper('Fetching and pulling updates...')
      self._sshClient.Call(f'cd {self._remoteProjectPath} && git clean -fd && git reset --hard && git checkout {self._gitBranch} && git fetch && git pull && git status')
    
    except IOError:
      self._PrintWrapper('Repository was not found on the remote!')
      self._PrintWrapper('Cloning repository (this may take a while!)...')
      self._sshClient.Call(f'cd {self._remoteProjectPath}')
      findKey = 'https://'
      index = self._gitURL.find(findKey) + len(findKey)
      url = f'{self._gitURL[:index]}{self._gitUsername}:{self._gitPassword}@{self._gitURL[index:]} {self._remoteProjectPath}'
      self._sshClient.Call(f'git clone -b {self._gitBranch} {url}')
      self._PrintWrapper('Finished cloning the repository!')


    # Close connection
    self._PrintWrapper('Closing SFTP connection...')
    self._sshClient.CloseSFTP(sftpClient)
    
  def CompileProject(self):
    self._PrintWrapper('Invoking GameMaker compiler...')
    buildCommand = f'"C:/ProgramData/GameMakerStudio2/Cache/runtimes/runtime-{self._runtimeVersion}/bin/Igor.exe"  -j=8 -options="{self.outputFolder}/Input/Options/build.bff" -v -- Windows PackageZip'
    self._sshClient.Call(buildCommand, printResult=self._verbose)
    print('Compiler has finished running.')

  def RetrieveBuild(self):
    sftpClient = self._sshClient.OpenSFTP()
    try:
      path = os.path.join(self.outputFolder, f'Output/GameZip/{self._projectName}.zip')
      sftpClient.stat(path)
      self._PrintWrapper('Sending the output back to local machine...')
      sftpClient.get(path, f'./{self._projectName}.zip')
      print(f'Output build is located at ./{self._projectName}.zip')
    except:
      self._PrintWrapper(colored('Failed to retrieve ouput file!', 'red'))
      


"""lazybuild: a remote compiling tool for GameMaker Studio 2"""

# Import libraries
import sys
import os
import subprocess
import json 
import time
import getpass
import base64
import aws
import ssh
import gmBuilder
from termcolor import colored
import colorama
colorama.init()

__author__ = "Gabe Weiner"
__license__ = "MIT"
__version__ = "1.0.2"
__maintainer__ = "Gabe Weiner"
__email__ = "imlazyeye (at) gmail.com"
__status__ = "Prototype"

class LazyBuild:

  def __init__(self):
    
    # Core data
    print('Loading user defined options...')
    self._configurationKeys = ['region', 'instanceID', 'sshUsername', 'sshPassword', 'yoyoID', 'runtimeVersion', 'steamSDKPath', 'yypPath', 'configuration', 'gitUsername', 'gitPassword', 'gitBranch', 'gitURL']
    try:
      self.config = self.LoadConfig()
    except:
      print(colored('Configure options were not found. Please set them now!', 'yellow'))
      self.config = {}
      for key in self._configurationKeys:
        self.config[key] = ''
      self.Configure()
    print('Checking SSH permission file...')
    self.sshKeyFilePath = './resources/misc/rsa.pem'
    if not os.path.exists(self.sshKeyFilePath):
      print(colored('SSH permission file not found. Please refer to the README!', 'red'))
      sys.exit()
    print('Initializing...')
    self.region = self.config['region']
    self.instanceID = self.config['instanceID']
    self.sshUsername = self.config['sshUsername']
    self.sshPassword = self.config['sshPassword']
    self.shutdownOnDestroy = True
    self.inputCommands = {
      'configure': 'Configure the needed options for lazybuild',
      'status': 'Check if the remote server is online',
      'startup': 'Starts the remote server',  
      'shutdown': 'Shuts down the remote server',
      'build': 'Builds the project on the remote server',
      'console': 'Connects you to your instance via SSH',
      'rdp': 'Starts a remote desktop session with the remote server',
      'help': 'Prints these commands again',
      'exit': 'Exits this program'
    }
    print('Fetching instance...')
    try:
      self.builderInstance = aws.AWSInstance(self.region, self.instanceID, self.shutdownOnDestroy)
    except:
      print(colored('There was an issue trying to access your remote machine!', 'red'))
      if self.YesNoPrompt('Would you like to edit your configuration options?'):
        self.Configure()
        print('Please start lazybuild again to apply these changes.')
      sys.exit()

    # Welcome message
    asciiMessage = r'''
-------------------------------------------------

      _                 _           _ _     _ 
    | |               | |         (_) |   | |
    | | __ _ _____   _| |__  _   _ _| | __| |
    | |/ _` |_  / | | | '_ \| | | | | |/ _` |
    | | (_| |/ /| |_| | |_) | |_| | | | (_| |
    |_|\__,_/___|\__, |_.__/ \__,_|_|_|\__,_|
                  __/ |                      
                  |___/                       
-------------------------------------------------
    '''
    print(colored(asciiMessage, 'green'))
    welcomeMessage = f'''Welcome to lazybuild! You are running version {__version__}

Please ensure you have read the README file before continuing!

Available commands:
    '''
    print(welcomeMessage)
    self.Help()

  def YesNoPrompt(self, message):
    resp = input(f'{message} [y/n] ')
    if (resp =='y'):
      return True 
    elif (resp == 'n'):
      return False 
    else:
      print(colored('Response invalid! Please use either "y" or "n".', 'yellow'))
      return self.YesNoPrompt(message)

  def Help(self):
    commands = ''
    for key, value in self.inputCommands.items():
      name = colored(key, 'cyan')
      commands += f'{name}: {value}\n'
    print(commands)

  def SaveConfig(self): # I just don't wanna save passwords in plaintext, man
    print('Saving configuration...')
    open('lazybuildConfig.dat', 'wb').write(base64.b64encode(json.dumps(self.config).encode('utf-8')))

  def LoadConfig(self):
    print('Opening configuration...')
    text = open('lazybuildConfig.dat', 'rb').read()
    return json.loads(base64.decodebytes(text))

  def ReadUserInput(self):
    commandToRun = input('>> ')
    try:
      self.inputCommands[commandToRun] # ensures this is an allowed function call
      getattr(self, commandToRun.capitalize())()
    except KeyError:
      print(colored('Command not recognized! Please try again.', 'yellow'))
      return

  def Configure(self):
    def InputHandle(key):
      currentValueString = '[]'
      if self.config[key] != '':
        currentValueString = f'[{self.config[key]}]'
      message = '{:<30}'.format(f'{key} {currentValueString}:')
      if 'password' in key.lower():
        resp = getpass.getpass(message)
      else:
        resp = input(message)
      if resp != '':
        self.config[key] = resp

    print('\nProvide a value for each option or press enter to keep the current value.')
    print('Please check the README for more information on each option!\n')
    for key in self._configurationKeys:
      InputHandle(key)
    self.SaveConfig()

  def Status(self):
    status = 'not online'
    if self.builderInstance.online:
      status = 'online'
    print(f'Remote machine is {status}.')

  def Console(self):
    if not self.builderInstance.online:
      print(colored('Remote machine is not online!', 'yellow'))
      return
    os.system(f'ssh -i {self.sshKeyFilePath} {self.sshUsername}@{self.builderInstance.GetInstanceIP()}')
    
  def Rdp(self):
    if not self.builderInstance.online:
      print(colored('Remote machine is not online!', 'yellow'))
      return
    open('temp.rdp', 'w+').write(f'auto connect:i:1\nfull address:s:{self.builderInstance.GetInstanceIP()}\nusername:s:{self.sshUsername}\ndrivestoredirect:s:')
    os.system('temp.rdp')
    time.sleep(1)
    os.remove('temp.rdp')
    print('Started remote desktop connection!')

  def Build(self):
    if not self.builderInstance.online:
      print(colored('Remote machine is not online!', 'yellow'))
      return
    else:
      clearCache = self.YesNoPrompt('Would you like to clear the cache prior to building?')
      verbose = self.YesNoPrompt('Would you like verbose output?')
      sshClient = ssh.SSHClient(self.sshUsername, self.builderInstance.GetInstanceIP(), self.sshPassword, self.sshKeyFilePath)
      if not verbose:
        print('Starting build (this may take a while!)...')
      myBuilder = gmBuilder.GMBuilder(self.config, sshClient, verbose)
      myBuilder.CreateBuildFolders(clearCache=clearCache)
      myBuilder.DumpOptionFiles()
      myBuilder.UpdateProject()
      myBuilder.CompileProject()
      myBuilder.RetrieveBuild()
      if self.YesNoPrompt('Would you like to shut down the instance?'):
        self.builderInstance.Shutdown()

  def Startup(self):
    self.builderInstance.Startup()
  
  def Shutdown(self):
    self.builderInstance.Shutdown()

  def Exit(self):
    sys.exit()

def main():
  lazyBuild = LazyBuild()
  while (True):
    lazyBuild.ReadUserInput()

# Run the script!
if __name__ == '__main__':
    main()